
import warnings
from typing import Any, List
from views_storage.serializers.json import JsonSerializable
from viewser.storage import model_object

class Storage():
    """
    Storage
    =======

    A storage driver which can be used to store and retrieve (model) objects
    and associated metadata. Works as a key-value store, with each (model)
    object being associated with a unique name, which is then used to fetch
    both the object and its metadata.

    """

    def __init__(self):
        self._model_object_store = model_object.ModelObjectStorage()
        self._metadata_store = model_object.ModelMetadataStorage()

    def store(self, name: str, model: Any, metadata: JsonSerializable = {}, overwrite: bool = False) -> None:
        """
        store
        =====

        parameters:
            name (str): Unique name associated with the model
            model (Any): A trained model object
            metadata (JsonSerializable): A dictionary containing metadata
            overwrite (bool) = False: Whether to overwrite the model if it already exists

        Store a model object and its associated metadata under a unique name.

        """
        self._model_object_store.write(name, model, overwrite = overwrite)
        self._metadata_store.write(name, metadata, overwrite = overwrite)

    def retrieve(self, name: str) -> Any:
        """
        retrieve
        ========

        parameters:
            name (str): Name of the model to fetch

        returns:
            Any: A previously stored model object.

        Retrieve a stored model object.
        """
        return self._model_object_store.read(name)

    def fetch_metadata(self, name: str) -> JsonSerializable :
        """
        retrieve
        ========

        parameters:
            name (str): Name of the model to fetch metadata for

        returns:
            JsonSerializable: Metadata for model 

        Retrieve previously stored metadata for a model.
        """
        return self._metadata_store.read(name)

    def list(self) -> List[str]:
        """
        list
        ====

        returns:
            List[str]

        Return a list of available model names.
        """
        return self._model_object_store.list()

def store(name: str, model: Any, metadata: Any):
    warnings.warn("This function is deprecated. Use the views_runs.Storage and its methods instead")
    storage = Storage()
    storage.store(name, model, metadata)

def retrieve(name: str):
    warnings.warn("This function is deprecated. Use the views_runs.Storage and its methods instead")
    storage = Storage()
    return storage.retrieve(name)

def fetch_metadata(name: str):
    warnings.warn("This function is deprecated. Use the views_runs.Storage and its methods instead")
    storage = Storage()
    return storage.fetch_metadata(name)

def list():
    warnings.warn("This function is deprecated. Use the views_runs.Storage and its methods instead")
    storage = Storage()
    return storage.list()




from stepshift.views import StepshiftedModels
from views_partitioning import DataPartitioner
import pandas as pd

from . import validation

class ViewsRun():
    """
    ViewsRun
    ========

    parameters:
        partitioner: views_partitioning.DataPartitioner
        models: stepshift.views.StepshiftedModels

    This class combines modelling and partitioning, delivering predictions in a
    familiar format and concise API.

    Takes two arguments, a views_partitioning.DataPartitioner instance, and a
    stepshift.StepshiftedModels instance. These instances determine how data is
    subset, and how modelling happens.
    """

    def __init__(self, partitioner: DataPartitioner, models: StepshiftedModels):
        self._models = models
        self._shifted_partitioner = partitioner.shift_left(self._models._steps_extent)
        self._partitioner = partitioner

    @validation.views_validate
    def fit(self, partition_name: str, timespan_name: str, data: pd.DataFrame)-> None:
        """
        fit
        ===

        parameters:
            partition_name (str)
            timespan_name (str)
            data (pandas.DataFrame)

        Fit the model to the named partition. partition_name must be present in
        the DataPartitioner.
        """

        self._models.fit(self._shifted_partitioner(partition_name,timespan_name,data))

    @validation.views_validate
    def predict(self, partition_name: str, timespan_name: str, data: pd.DataFrame)-> pd.DataFrame:
        """
        predict
        =======

        parameters:
            partition_name (str)
            timespan_name (str)
            data (pandas.DataFrame)

        Predict with data from the named partition. partition_name must be
        present in the DataPartitioner. Returns predictions within the selected
        timespan.
        """

        predictions = self._models.predict(
                self._shifted_partitioner(partition_name, timespan_name, data),
                combine = False)
        predictions.index.names = data.index.names
        data = data.merge(predictions, how = "left", left_index = True, right_index = True)
        return self._partitioner(partition_name,timespan_name,data)

    @validation.views_validate
    def future_predict(self, partition_name: str, timespan_name: str, data: pd.DataFrame, keep_specific: bool = False) -> pd.DataFrame:
        """
        future_predict
        ==============

        parameters:
            partition_name (str)
            timespan_name (str)
            data (pandas.DataFrame)
            keep_specific (bool) = False: Return step-specific predictions as well

        returns:
            pandas.DataFrame: Dataframe containing future predictions

        Predict into the future using data from the last point in time in the
        selected partition. Returns step-combined predictions.
        """

        predictions = self._models.predict(
                self._shifted_partitioner(partition_name, timespan_name, data),
                combine = True)
        end = self._shifted_partitioner.partitions.partitions[partition_name].timespans[timespan_name].end + 1
        future_preds = predictions.loc[end:]
        if not keep_specific:
            future_preds = future_preds[["step_combined"]]

        return future_preds

    def future_point_predict(self, time: int, data: pd.DataFrame, keep_specific: bool = False) -> pd.DataFrame:
        """
        future_point_predict 
        ====================

        parameters:
            time (int): Point in time to predict from, given in views-month (months since 1979-12)
            data (pandas.DataFrame): Data to predict from
            keep_specific (bool) = False: Return step-specific predictions as well

        returns:
            pandas.DataFrame

        Predict into the future from a point in time.
        """
        predictions = self._models.predict(
                data.loc[time - self._models._steps_extent: time],
                combine = True
                )

        if not keep_specific:
            predictions = predictions[["step_combined"]]
        

        return predictions.loc[time+1 : time + self._models._steps_extent]

    @property
    def models(self):
        return self._models.models
